
# 单变量梯度下降

本讲座/实验的目标是解决一个重要的迭代计算问题：*梯度下降函数最小化*。 我希望你会看到梯度下降与平方根近似相同，只是具有不同的递归关系。 本课程的完整[源代码在这里](https://github.com/parrt/msan501/blob/master/notes/code/descent.py)。

## 讨论

找到最小化或“优化”函数`f(x)`（通常在某个范围内）的`x`是一项非常重要的操作，因为我们可以使用它来最小化风险，并且对于机器学习，可以学习我们的分类器或预测器的参数。 例如，深度学习使用成本函数优化来训练神经网络。 通常`x`将是一个向量，但我们假设`x`是一个标量来学习基础知识。 如果我们知道函数像二次多项式一样是凸的，那么有一个独特的解决方案，我们可以简单地将导数设置为零并求解`x`：

```
f'(x) = 0
```

这基本上是在问“这个函数在哪里变平？” 例如，函数`f(x) = (x-2)^2 + 1`的导数为`f'(x) = 2x - 4`，其零点为`x = 2`。

```python
import matplotlib.pyplot as plt
import numpy as np

def f(x): return (x - 2)**2 + 1

graphx = np.arange(0, 4, 0.01)
graphy = f(graphx) # apply f to all of graphx (BROADCAST)
plt.plot(graphx, graphy)
plt.show()
```

![png](img/4.3_gradient-descent_3_0.png)

```python
def foo(x): 
    print(type(x))
    return (x - 2)**2 + 1
foo(2), foo(10)

'''
<class 'int'>
<class 'int'>

(1, 65)
'''
```

```python
X = np.array([1,2,3])
foo(X)

'''
<class 'numpy.ndarray'>

array([2, 1, 2])
'''
```

我们更喜欢找到`f(x)`的*全局最小值*，但通常必须满足*局部最小值*，我们希望它接近全局最小值。 找到全局最小值的一个不错的方法是通过随机初始化`x0`的位置，找到一些局部最小值，然后选择所发现的`f(x)`最小的最小值的`x`。 例如，函数`f(x) = cos(3\pi x) / x`在`[0,1.3]`中有两个最小值，其中一个是明显的全局最小值：

```python
def f(x): return np.cos(3 * np.pi * x) / x

graphx = np.arange(.1, 1.1, 0.01)
graphy = f(graphx)
plt.plot(graphx, graphy)
plt.axis([0, 1.1, -4, 6])
plt.show()
```


![png](img/4.3_gradient-descent_7_0.png)

如果函数有很多最小值/最大值或非常复杂，则可能没有简单的解析解。

有许多方法可以迭代地（即非解析地）找到函数最小值，但我们将使用一种众所周知的技术，称为*梯度下降*或*最速下降法*。

### 梯度下降

梯度下降需要一个起始位置，`x0`，优化函数，`f(x)`及其导数`f'(x)`。 回想一下，导数只是特定点处函数的斜率。 换句话说，当`x`从特定位置移动时，`f(x)`是上涨还是下跌，以及涨多少？ 例如，`x ^ 2`的导数是`2x`，当`x > 0`时为正斜率，当`x < 0`时为负斜率。 梯度下降使用导数，迭代地选择`x`的新值，使我们越来越接近`f(x)`的最小值。 斜率的符号告诉我们最接近的最小值的方向（以`x`为单位）。 例如，这里再次是余弦图，这次显示了代表特定点的导数的多个向量。

<img src="img/cos-2minima-edited.png" width="300">

注意，导数在最小值处为零，即平坦（对于最大值也是如此）。 请注意，导数（红色向量）指向最接近的最小值的相反方向。 如果我们想要朝最近的最小值方向移动，我们应该通过加上导数的相反数来调整`x`; 也就是说，我们应该从`xi`中减去导数，得到`x_{i + 1}`。 更新我们对`x`的估计来最小化`f(x)`的递归关系就是：

```
x_{i+1} = xi - η f'(xi)
```

其中`η`被称为*学习率*，我们将在下面讨论。` η f'(xi)`一项表示我们采取的最小步骤的大小。

基本的算法为：

+   选取初始值`x0`，让`x = x0`
+   让`x_{i+1} = xi - η f'(x_i)` 直到`f'(x_i)=0`

该算法非常简单，但在处理计算机的有限精度时，知道何时停止算法是个问题。 具体来说，没有两个浮点数真的相等。所以`f'(x) = 0`总是假的。通常我们会这样做：

```
abs(x_{i+1} - xi) < precision
```

精度是一些非常小的数字，如 0.0000001。 我们也可以测试一下

```
abs(f(x_{i+1}) - f(xi)) < precision
```

看看函数是否变平或者是否有一个非常小的垂直变化，**和**`f(x_{i+1})`是否正在回升。第三种方法是测试导数是否接近于零。 我尝试了所有这三个，并发现只检查`x_ {i + 1}`是否收敛是最好的（例如，它允许我具有非常高的学习率）。 第三种方法是有问题的，因为我们使用有限差分，这是对真实导数的不良近似。

我们采取的步骤按学习率`η`缩放。 Yaser S. Abu-Mostafa 有一些[很棒的幻灯片](http://www.amlbook.com/slides/iTunesU_Lecture09_May_01.pdf)和你应该看的视频。以下是他在幻灯片 21 中对学习率如何影响收敛的描述：

<img src="img/stepsize.png" width="400">

The domain of $x$ also affects the learning rate magnitude. This is all a very complicated finicky business and those experienced in the field tell me it's very much an art picking the learning rate, starting positions, precision, and so on. You can start out with a low learning rate and crank it up to see if you still converge without oscillating around the minimum.  An excellent description of gradient descent and other minimization techniques can be found in [Numerical Recipes](http://apps.nrbook.com/fortran/index.html).

### Approximating derivatives with finite differences

Sometimes, the derivative is hard, expensive, or impossible to find analytically (symbolically).  For example, some functions are themselves iterative in nature or even simulations that must be optimized. There might be no closed form for $f(x)$. To get around this and to reduce the input requirements, we can approximate the derivative in the neighborhood of a particular $x$ value. That way we can optimize any reasonably well behaved function (left and right continuity would be nice). Our minimizer then only requires a starting location and $f(x)$ but not $f'(x)$, which makes the lives of our users much simpler and our minimizer much more flexible. 

To approximate the derivative, we can take several approaches. The simplest involves a comparison. Since we really just need a direction, all we have to do is compare the current $f(x_i)$ with values a small step, $h$, away in either direction: $f(x_{i}-h)$ and $f(x_{i}+h)$.  If $f(x_{i}-h) < f(x_{i})$, we should move $x_{i+1}$ to the left of $x_{i}$. If $f(x_{i}+h) < f(x_{i})$, we should move $x_{i+1}$ to the right.  These are called the backward and forward differences, but there is also a central difference. The excellent article [Stochastic Gradient Descent Tricks](http://research.microsoft.com/pubs/192769/tricks-2012.pdf) has a lot of practical information on computing gradients etc...

Using the direction of the slope works, but does not converge very fast. What we really want is to use the magnitude of the slope to make the algorithm go fast where it's steep and slow where it's shallow because it will be approaching a minimum. So, rather than just using the sign of the finite difference, we should use the magnitude or rate of change. Replacing the derivative in our recurrence relation with the finite (forward) difference, we get a similar formula:

\\[
x _{i+1} = x_i - \eta \frac{f(x_{i}+h) - f(x_{i})}{h} \text{ where } f'(x) \approx \frac{f(x_{i}+h) - f(x_{i})}{h}
\\]

To simplify things, we can roll the step size $h$ into the learning rate $\eta$ constant as we are going to pick that anyway.

\\[
x _{i+1} = x_i - \eta (f(x_{i}+h) - f(x_{i}))
\\]

 The step size is bigger when the slope is bigger and is smaller as we approach the minimum (since the region is flatter). Abu-Mostafa indicates in his slides that $\eta$ should increase with the slope whereas we are keeping it fixed and allowing the finite difference to increase the step size. We are not normalizing the derivative/difference to a unit vector like he does (see his slides).

## An implementation

Our goal is to use gradient descent to minimize $f(x) = cos(3\pi x) / x$. To increase chances of finding the global minimum, we can pick a few random starting locations in the range $[0.1,1.3]$ using standard python {\tt random.uniform()} and perform gradient descent on all of them. To  observe our minimizer in action, we'll eventually plot the trace of $x$'s that indicate the steps taken by our gradient descent. Here are two sample descents where the $x$ and $f(x)$ values are displayed as well as the minima:

<img src="img/cos-trace-2minima.svg" width="350">

<img src="img/cos-trace-2minima-another.svg" width="350">

Recall from our square root lecture that we had a basic outline for an iterative method:

```python
x_prev = initial value
while True:
    x_next = function-giving-next-value(x_prev)
    if abs(x_next - x_prev) < precision:
        return x_next
```

To implement our gradient descent, we just replace function-giving-next-value with $x_i - \eta (f(x_{i}+h) - f(x_{i}))$:


```python
def minimize(f, x0, eta, h, precision):
    x = x0
    while True:
        prev = x
        finite_diff = f(x + h) - f(x)	 # division by h rolls into learning rate
        x = x - eta * finite_diff 		# decelerates x step as it flattens out
        if abs(x - prev) < precision:
            return x   # x is coordinate where f(x) is a minimum
```

Choosing an appropriate step value $h$ is important to get a decent approximation of the derivative through finite differences but that is large enough to avoid faulty results from lack of precision (subtracting two floating-point numbers in the computer results in a number with much less precision than the original numbers). You want that number to be small enough so that your algorithm does not oscillate around the minimum. If $\eta$ is too big the recurrence relation will compute a finite difference that makes $x_{i+1}$ leap across the minimum to the other wall of the function.  We must also pick a learning rate $\eta$ that allows us to go as fast as we can but not so fast that $x_{i+1}$ overshoots the minimum back and forth. When I crank up my learning rate too far, I see the algorithm oscillate:

```
...
f(2.470932790352) = -0.109488272790 , delta = -0.39601471581233022023
f(2.099265953523) = 0.282655040497 , delta = 0.39214331328753893047
f(2.474216657712) = -0.097250083113 , delta = -0.37990512360965078553
f(2.100529851383) = 0.277900324077 , delta = 0.37515040718930170449
f(2.478172114663) = -0.082429682649 , delta = -0.36033000672594917013
...
```

Here's are some decent parameters and how we can call the minimize function:


```python
import random
ETA = 10
h = 0.0001
PRECISION = 0.0000001 # can't be too small as f(x)-f(xprev) prec is low
def f(x): return np.cos(3*np.pi*x) / x
x0 = random.uniform(.1, 1.2)
minx = minimize(f, x0, ETA, h, PRECISION)
print(f"f({minx}) = {f(minx)}")

# f(0.9886073570106972) = -1.0056986067028884
```


To help you understand what the program is doing, you can print out $x$, $f(x)$, and any other value you think is helpful to see how your program explores the curve. This goes inside the loop:

```python
print(f"f({x:.12f}) = {f(x):.12f} delta = {delta:.20f}")
```

## Displaying the minimization trace

Instead of just printing the values out, it's much easier to understand the path taken by the minimization function if we visualize it. First thing we need is a version of the function that keeps a trace of all $x$ values:


```python
def minimize_trace(f, x0, eta, h, precision):
    tracex = [x0]
    x = x0
    while True:
        prev = x
        finite_diff = f(x + h) - f(x)  # /h rolls into learning rate
        x = x - eta * finite_diff  # decelerates x step as it flattens out
        tracex.append(x)
        if abs(x - prev) < precision:
            return tracex
```

Next, we import the matplotlib library for plotting and then define the following display function.


```python
import matplotlib.pyplot as plt

def viz_trace(x0, minimizer):
    # Plot the damped sine curve
    graphx = np.arange(.1, 1.3, 0.01)
    graphy = [f(x) for x in graphx] # or just f(graphx)!
    plt.plot(graphx, graphy)

    # Minimize and get trace of x locations
    tracex = minimizer(f, x0, ETA, h, PRECISION)

    # Plot the trace
    tracey = [f(x) for x in tracex]
    plt.scatter(tracex, tracey, color="darkred", marker='o', s=5)
    
    # Add some text describing trace
    plt.text(0.3, 4.5, f"f(x={tracex[-1]:.5f}) = {f(tracex[-1]):.5f}",
             fontsize=14)
    plt.text(0.3, 3.7, f"steps = {len(tracex)}", fontsize=14)
    plt.tight_layout()
    plt.savefig("img/cos-trace-2minima-another.svg")
    plt.show()
```

To test that routine, we pass in some random starting point and let it display the trace:


```python
# every time you run this, it'll start at new initial location and find nearest minimum
import time
random.seed(int(round(time.time() * 1000))) # get new pseudo-random sequence each time
x0 = random.uniform(.1, 1.3)
viz_trace(x0, minimize_trace)
```


![png](img/4.3_gradient-descent_23_0.png)


# Using symbolic derivatives

Instead of the finite difference solution, we can do a better job using the [symbolic derivative](https://m.wolframalpha.com/input/?i=derivative+cos%283*pi*x%29+%2F+x), which we can get for free for any function using Wolfram Alpha or [other tools](https://www.derivative-calculator.net/#expr=cos%283%2Api%2Ax%29%20%2F%20x&showsteps=1).  Finite difference is fine in theory but they don't work as well with finite precision floating point numbers (i.e., on a computer). Subtraction destroys precision, particularly when the operands are very similar.

\\[
\frac{d}{dx} \frac{cos(3\pi x)}{x} = - \frac{3 \pi x sin(3 \pi x) + cos(3 \pi x)}{x^2}
\\]


```python
def minimize_dx(f, x0, eta, h, precision):
    tracex = [x0]
    x = x0
    while True:
        prev = x
        dx = - (3 * np.pi * x * np.sin(3*np.pi*x) + np.cos(3 * np.pi * x)) / x**2
        x = x - eta * dx 		# decelerates x step as it flattens out
#         if len(tracex)<300:
#             print(f"x={x:.3f}, dx={dx:.3f}")
        tracex.append(x)
        if abs(x - prev) < precision:
            return tracex
```


```python
ETA = 0.001
random.seed(int(round(time.time() * 1000))) # get new pseudo-random sequence each time
x0 = random.uniform(.1, 1.3)
viz_trace(x0, minimize_dx)
```


![png](img/4.3_gradient-descent_26_0.png)


## Summary

Lots of mathematical computing problems require an iterative solution. One of the most important is minimizing error functions to train machine learning models. The good news is that the gradient descent method described here follows the simple iterative method outline we saw when approximating square root. The next step is to extend gradient descent to two variables, which you will do as a project to solve a regression (line fitting) problem.
